import logging

import click

from copernicus_marine_client.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicus_marine_client.command_line_interface.utils import (
    MutuallyExclusiveOption,
    tqdm_disable_option,
)
from copernicus_marine_client.core_functions.describe import describe_function

logger = logging.getLogger("copernicus_marine_root_logger")
blank_logger = logging.getLogger("copernicus_marine_blank_logger")


@click.group()
def cli_group_describe() -> None:
    pass


@cli_group_describe.command(
    "describe",
    short_help="Print Copernicus Marine catalog as JSON.",
    help="""
    Print Copernicus Marine catalog as JSON.

    The default display contains information on the products, and more data
    can be displayed using the --include-<argument> flags.

    The --contains option allows the user to specify one or several strings to
    filter through the catalogue display. The search is performed recursively
    on all attributes of the catalogue, and the tokens only need to be
    contained in one of the attributes (i.e. not exact match).
    """,
    epilog="""
    Examples:

    \b
    copernicus-marine describe --contains METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2 --include-datasets

    \b
    copernicus-marine describe -c METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2
    """,  # noqa
)
@click.option(
    "--include-description",
    type=bool,
    is_flag=True,
    default=False,
    help="Include product description in output.",
)
@click.option(
    "--include-datasets",
    type=bool,
    is_flag=True,
    default=False,
    help="Include product dataset details in output.",
)
@click.option(
    "--include-keywords",
    type=bool,
    is_flag=True,
    default=False,
    help="Include product keyword details in output.",
)
@click.option(
    "--contains",
    "-c",
    type=str,
    multiple=True,
    help="Filter catalogue output. Returns products with attributes "
    "matching a string token.",
)
@click.option(
    "--overwrite-metadata-cache",
    cls=MutuallyExclusiveOption,
    type=bool,
    is_flag=True,
    default=False,
    help="Force to refresh the catalogue by overwriting the local cache.",
    mutually_exclusive=["no_metadata_cache"],
)
@click.option(
    "--no-metadata-cache",
    cls=MutuallyExclusiveOption,
    type=bool,
    is_flag=True,
    default=False,
    help="Bypass the use of cache.",
    mutually_exclusive=["overwrite_metadata_cache"],
)
@tqdm_disable_option
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=(
        "Set the details printed to console by the command "
        "(based on standard logging library)."
    ),
)
@click.option(
    "--staging",
    type=bool,
    default=False,
    is_flag=True,
    hidden=True,
)
@log_exception_and_exit
def describe(
    include_description: bool,
    include_datasets: bool,
    include_keywords: bool,
    contains: list[str],
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    disable_progress_bar: bool,
    log_level: str,
    staging: bool,
) -> None:
    if log_level == "QUIET":
        logger.disabled = True
        logger.setLevel(level="CRITICAL")
    else:
        logger.setLevel(level=log_level)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("DEBUG mode activated")

    json_dump = describe_function(
        include_description=include_description,
        include_datasets=include_datasets,
        include_keywords=include_keywords,
        contains=contains,
        overwrite_metadata_cache=overwrite_metadata_cache,
        no_metadata_cache=no_metadata_cache,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )
    blank_logger.info(json_dump)



import fnmatch
import json
import logging
import pathlib
from typing import List, Optional

from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineDatasetServiceType,
    parse_catalogue,
)
from copernicus_marine_client.catalogue_parser.request_structure import (
    GetRequest,
    get_request_from_file,
)
from copernicus_marine_client.core_functions.credentials_utils import (
    get_and_check_username_password,
)
from copernicus_marine_client.core_functions.services_utils import (
    CommandType,
    get_dataset_service_and_suffix_path,
)
from copernicus_marine_client.core_functions.utils import get_unique_filename
from copernicus_marine_client.core_functions.versions_verifier import (
    VersionVerifier,
)
from copernicus_marine_client.download_functions.download_ftp import (
    download_ftp,
)
from copernicus_marine_client.download_functions.download_original_files import (
    download_original_files,
)

logger = logging.getLogger("copernicus_marine_root_logger")


def get_function(
    dataset_url: Optional[str],
    dataset_id: Optional[str],
    username: Optional[str],
    password: Optional[str],
    no_directories: bool,
    show_outputnames: bool,
    output_directory: Optional[pathlib.Path],
    credentials_file: Optional[pathlib.Path],
    force_download: bool,
    overwrite_output_data: bool,
    request_file: Optional[pathlib.Path],
    force_service: Optional[str],
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    filter: Optional[str],
    regex: Optional[str],
    disable_progress_bar: bool,
    staging: bool,
) -> List[pathlib.Path]:
    VersionVerifier.check_version_get(staging)
    if staging:
        logger.warning(
            "Detecting staging flag for get command. "
            "Data will come from the staging environment."
        )

    get_request = GetRequest()
    if request_file:
        get_request = get_request_from_file(request_file)
    request_update_dict = {
        "dataset_url": dataset_url,
        "dataset_id": dataset_id,
        "output_directory": output_directory,
        "force_service": force_service,
    }
    get_request.update(request_update_dict)

    # Specific treatment for default values:
    # In order to not overload arguments with default values
    if no_directories:
        get_request.no_directories = no_directories
    if show_outputnames:
        get_request.show_outputnames = show_outputnames
    if force_download:
        get_request.force_download = force_download
    if overwrite_output_data:
        get_request.overwrite_output_data = overwrite_output_data
    if force_service:
        get_request.force_service = force_service
    if filter:
        get_request.regex = _filter_to_regex(filter)
    if regex:
        get_request.regex = _overload_regex_with_filter(regex, filter)

    return _run_get_request(
        username,
        password,
        get_request,
        credentials_file,
        overwrite_metadata_cache,
        no_metadata_cache,
        disable_progress_bar,
        staging=staging,
    )


def _filter_to_regex(filter: str) -> str:
    return fnmatch.translate(filter)


def _overload_regex_with_filter(regex: str, filter: Optional[str]) -> str:
    return (
        "(" + regex + "|" + _filter_to_regex(filter) + ")" if filter else regex
    )


def _run_get_request(
    username: Optional[str],
    password: Optional[str],
    get_request: GetRequest,
    credentials_file: Optional[pathlib.Path],
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    disable_progress_bar: bool,
    staging: bool = False,
) -> List[pathlib.Path]:
    username, password = get_and_check_username_password(
        username,
        password,
        credentials_file,
    )
    catalogue = parse_catalogue(
        overwrite_metadata_cache=overwrite_metadata_cache,
        no_metadata_cache=no_metadata_cache,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )
    dataset_service, _, suffix_path = get_dataset_service_and_suffix_path(
        catalogue,
        get_request.dataset_id,
        get_request.dataset_url,
        get_request.force_service,
        CommandType.GET,
    )
    get_request.dataset_url = (
        dataset_service.uri + suffix_path
        if suffix_path
        else dataset_service.uri
    )
    logger.info(
        "Downloading using service "
        f"{dataset_service.service_type.service_name.value}..."
    )
    downloaded_files = (
        download_ftp(
            username,
            password,
            get_request,
            disable_progress_bar,
        )
        if dataset_service.service_type
        == CopernicusMarineDatasetServiceType.FTP
        else download_original_files(
            username, password, get_request, disable_progress_bar
        )
    )
    logger.debug(downloaded_files)
    return downloaded_files


def create_get_template() -> None:
    filename = get_unique_filename(
        filepath=pathlib.Path("get_template.json"), overwrite_option=False
    )
    with open(filename, "w") as output_file:
        json.dump(
            {
                "dataset_url": (
                    "ftp://my.cmems-du.eu/Core/"
                    "IBI_MULTIYEAR_PHY_005_002/"
                    "cmems_mod_ibi_phy_my_0.083deg-3D_P1Y-m"
                ),
                "filter": "*01yav_200[0-2]*",
                "regex": False,
                "output_directory": "copernicusmarine_data",
                "show_outputnames": True,
                "force_service": "files",
                "force_download": False,
                "request_file": False,
                "overwrite_output_data": False,
                "overwrite_metadata_cache": False,
                "no_metadata_cache": False,
            },
            output_file,
            indent=4,
        )
    logger.info(f"Template created at: {filename}")



import datetime
import pathlib
import re
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import List, Optional, Tuple

import boto3
import botocore
from numpy import append, arange
from tqdm import tqdm

from copernicus_marine_client.catalogue_parser.request_structure import (
    GetRequest,
)
from copernicus_marine_client.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
    construct_url_with_query_params,
    flatten,
    get_unique_filename,
)
from copernicus_marine_client.download_functions.download_get import (
    download_get,
)

# Some buckets do not support listing. While these buckets are being replaced
# by other ones that are listable, we need to recognise the non-listable ones
# LIST via CDN.
MARINE_DATA_LAKE_CDN_LISTING_ENDPOINT = "https://marine.copernicus.eu"
MARINE_DATA_LAKE_NON_LISTABLE_BUCKETS = {
    "mdl-native": {
        "endpoint_url": MARINE_DATA_LAKE_CDN_LISTING_ENDPOINT,
        "bucket": "mdl-native-list",
    },
    "mdl-native-dta": {
        "endpoint_url": MARINE_DATA_LAKE_CDN_LISTING_ENDPOINT,
        "bucket": "mdl-native-list-dta",
    },
}


def download_original_files(
    username: str,
    password: str,
    get_request: GetRequest,
    disable_progress_bar: bool,
) -> list[pathlib.Path]:
    filenames_in, filenames_out, locator = download_get(
        username,
        password,
        get_request,
        _download_header,
        create_filenames_out,
    )
    endpoint: str
    bucket: str
    endpoint, bucket = locator
    return download_files(
        username,
        endpoint,
        bucket,
        filenames_in,
        filenames_out,
        disable_progress_bar,
    )


def download_files(
    username: str,
    endpoint_url: str,
    bucket: str,
    filenames_in: List[str],
    filenames_out: List[pathlib.Path],
    disable_progress_bar: bool,
) -> list[pathlib.Path]:
    pool = ThreadPool()
    nfiles_per_process, nfiles = 1, len(filenames_in)
    indexes = append(
        arange(0, nfiles, nfiles_per_process, dtype=int),
        nfiles,
    )
    groups_in_files = [
        filenames_in[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]
    groups_out_files = [
        filenames_out[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]

    for groups_out_file in groups_out_files:
        parent_dir = Path(groups_out_file[0]).parent
        if not parent_dir.is_dir():
            pathlib.Path.mkdir(parent_dir, parents=True)

    download_summary_list = pool.imap(
        _download_files,
        zip(
            [username] * len(groups_in_files),
            [endpoint_url] * len(groups_in_files),
            [bucket] * len(groups_in_files),
            groups_in_files,
            groups_out_files,
        ),
    )
    download_summary = list(
        tqdm(
            download_summary_list,
            total=len(groups_in_files),
            disable=disable_progress_bar,
        )
    )
    return flatten(download_summary)


def _download_header(
    data_path: str, regex: Optional[str], username: str, _password: str
) -> Tuple[str, Tuple[str, str], list[str], float]:
    (endpoint_url, bucket, path) = parse_original_files_dataset_url(data_path)
    filenames, sizes, total_size = [], [], 0.0
    raw_filenames = _list_files_on_marine_data_lake_s3(
        username, endpoint_url, bucket, path
    )
    filename_filtered = []
    for filename, size, last_modified_datetime in raw_filenames:
        if not regex or re.match(regex, filename):
            filenames += [filename]
            sizes += [float(size)]
            total_size += float(size)
            filename_filtered.append((filename, size, last_modified_datetime))

    message = "You requested the download of the following files:\n"
    for filename, size, last_modified_datetime in filename_filtered[:20]:
        message += str(filename)
        datetime_iso = re.sub(
            r"\+00:00$",
            "Z",
            last_modified_datetime.astimezone(datetime.timezone.utc).isoformat(
                timespec="seconds"
            ),
        )
        message += f" - {format_file_size(float(size))} - {datetime_iso}\n"
    if len(filenames) > 20:
        message += f"Printed 20 out of {len(filenames)} files\n"
    message += (
        f"\nTotal size of the download: {format_file_size(total_size)}\n\n"
    )
    locator = (endpoint_url, bucket)
    return (message, locator, filenames, total_size)


def _list_files_on_marine_data_lake_s3(
    username: str,
    endpoint_url: str,
    bucket: str,
    prefix: str,
) -> list[tuple[str, int, datetime.datetime]]:
    def _add_custom_query_param(params, context, **kwargs):
        """
        Add custom query params for MDS's Monitoring
        """
        params["url"] = construct_url_with_query_params(
            params["url"],
            construct_query_params_for_marine_data_store_monitoring(username),
        )

    # For non-listable buckets, use the alternatives (CDN)
    original_bucket = bucket
    if bucket in MARINE_DATA_LAKE_NON_LISTABLE_BUCKETS:
        alternative = MARINE_DATA_LAKE_NON_LISTABLE_BUCKETS[bucket]
        bucket = alternative["bucket"]
        endpoint_url = alternative["endpoint_url"]

    s3_session = boto3.session.Session()
    s3_client = s3_session.client(
        "s3",
        config=botocore.config.Config(
            # Configures to use subdomain/virtual calling format.
            s3={"addressing_style": "virtual"},
            signature_version=botocore.UNSIGNED,
        ),
        endpoint_url=endpoint_url,
    )

    # Register the botocore event handler for adding custom query params
    # to S3 LIST requests
    s3_client.meta.events.register(
        "before-call.s3.ListObjects", _add_custom_query_param
    )

    paginator = s3_client.get_paginator("list_objects")
    page_iterator = paginator.paginate(Bucket=bucket, Prefix=prefix)

    s3_objects = chain(*map(lambda page: page["Contents"], page_iterator))

    files_already_found = []
    for s3_object in s3_objects:
        files_already_found.extend(
            [
                (
                    f"s3://{original_bucket}/" + s3_object["Key"],
                    s3_object["Size"],
                    s3_object["LastModified"],
                )
            ]
        )
    return files_already_found


def _download_files(
    tuple_original_files_filename: Tuple[
        str, str, str, list[str], list[pathlib.Path]
    ],
) -> list[pathlib.Path]:
    (
        username,
        endpoint_url,
        bucket,
        filenames_in,
        filenames_out,
    ) = tuple_original_files_filename

    def _add_custom_query_param(params, context, **kwargs):
        """
        Add custom query params for MDS's Monitoring
        """
        params["url"] = construct_url_with_query_params(
            params["url"],
            construct_query_params_for_marine_data_store_monitoring(username),
        )

    def _original_files_file_download(
        endpoint_url: str, bucket: str, file_in: str, file_out: pathlib.Path
    ) -> pathlib.Path:
        """
        Download ONE file and return a string of the result
        """
        s3_session = boto3.session.Session()
        s3_client = s3_session.client(
            "s3",
            config=botocore.config.Config(
                # Configures to use subdomain/virtual calling format.
                s3={"addressing_style": "virtual"},
                signature_version=botocore.UNSIGNED,
            ),
            endpoint_url=endpoint_url,
        )

        # Register the botocore event handler for adding custom query params
        # to S3 HEAD and GET requests
        s3_client.meta.events.register(
            "before-call.s3.HeadObject", _add_custom_query_param
        )
        s3_client.meta.events.register(
            "before-call.s3.GetObject", _add_custom_query_param
        )

        s3_client.download_file(
            bucket,
            file_in.replace(f"s3://{bucket}/", ""),
            file_out,
        )

        return file_out

    download_summary = []
    for file_in, file_out in zip(filenames_in, filenames_out):
        download_summary.append(
            _original_files_file_download(
                endpoint_url, bucket, file_in, file_out
            )
        )
    return download_summary


# /////////////////////////////
# --- Tools
# /////////////////////////////


# Example data_path
# https://s3.waw3-1.cloudferro.com/mdl-native-01/native/NWSHELF_MULTIYEAR_BGC_004_011/cmems_mod_nws_bgc-pft_myint_7km-3D-diato_P1M-m_202105
def parse_original_files_dataset_url(data_path: str) -> Tuple[str, str, str]:
    match = re.search(r"^(http|https):\/\/([\w\-\.]+)(\/.*)", data_path)
    if match:
        endpoint_url = match.group(1) + "://" + match.group(2)
        full_path = match.group(3)
        segments = full_path.split("/")
        bucket = segments[1]
        path = "/".join(segments[2:])
        return endpoint_url, bucket, path
    else:
        raise Exception(f"Invalid data path: {data_path}")


def create_filenames_out(
    filenames_in: list[str],
    overwrite: bool,
    output_directory: pathlib.Path = pathlib.Path("."),
    no_directories=False,
) -> list[pathlib.Path]:
    filenames_out = []
    for filename_in in filenames_in:
        filename_out = output_directory
        if no_directories:
            filename_out = (
                pathlib.Path(filename_out) / pathlib.Path(filename_in).name
            )
        else:
            # filename_in: s3://mdl-native-xx/native/<product-id>...
            filename_out = filename_out / pathlib.Path(
                "/".join(filename_in.split("/")[4:])
            )

        filename_out = get_unique_filename(
            filepath=filename_out, overwrite_option=overwrite
        )

        filenames_out.append(filename_out)
    return filenames_out


def format_file_size(
    size: float, decimals: int = 2, binary_system: bool = False
) -> str:
    if binary_system:
        units: list[str] = [
            "B",
            "KiB",
            "MiB",
            "GiB",
            "TiB",
            "PiB",
            "EiB",
            "ZiB",
        ]
        largest_unit: str = "YiB"
        step: int = 1024
    else:
        units = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB"]
        largest_unit = "YB"
        step = 1000

    for unit in units:
        if size < step:
            return ("%." + str(decimals) + "f %s") % (size, unit)
        size /= step

    return ("%." + str(decimals) + "f %s") % (size, largest_unit)



import logging

import requests
import semver

import copernicus_marine_client
from copernicus_marine_client.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
)

logger = logging.getLogger("copernicus_marine_root_logger")


class VersionVerifier:
    function_marine_data_store_service_mapping: dict[str, list[str]] = {
        "describe": ["mds", "mds/serverlessArco/meta"],
        "get": ["mds", "mds/serverlessNative", "mds/serverlessArco/meta"],
        "subset": ["mds", "mds/serverlessArco", "mds/serverlessArco/meta"],
    }

    @staticmethod
    def check_version_describe(staging: bool):
        VersionVerifier._check_version("describe", staging)

    @staticmethod
    def check_version_get(staging: bool):
        VersionVerifier._check_version("get", staging)

    @staticmethod
    def check_version_subset(staging: bool):
        VersionVerifier._check_version("subset", staging)

    @staticmethod
    def _check_version(function_name: str, staging: bool):
        marine_data_store_versions = (
            VersionVerifier._get_client_required_versions(staging)
        )
        client_version = copernicus_marine_client.__version__
        for (
            service
        ) in VersionVerifier.function_marine_data_store_service_mapping[
            function_name
        ]:
            required_version = marine_data_store_versions[service]
            if not semver.match(client_version, required_version):
                logger.debug(
                    f"Client version {client_version} is not compatible with "
                    f"{service}. Service needs version {required_version}."
                )
                logger.error(
                    f"Client version {client_version} is not compatible with current "
                    "backend service. Please update to the latest client version."
                )

    @staticmethod
    def _get_client_required_versions(
        staging: bool,
    ) -> dict[str, str]:
        url_mds_versions = (
            "https://stac-dta.marine.copernicus.eu/mdsVersions.json"
            if staging
            else "https://stac.marine.copernicus.eu/mdsVersions.json"
        )
        logger.debug(f"Getting required versions from {url_mds_versions}")
        mds_versions: dict[str, str] = requests.get(
            url_mds_versions,
            params=construct_query_params_for_marine_data_store_monitoring(),
        ).json()["clientVersions"]
        return mds_versions



import logging
import pathlib

import xarray
import zarr
from dask.delayed import Delayed
from tqdm.dask import TqdmCallback

logger = logging.getLogger("copernicus_marine_root_logger")


def get_delayed_download(dataset: xarray.Dataset, output_path: pathlib.Path):
    if output_path.suffix == ".nc":
        delayed = _prepare_download_dataset_as_netcdf(dataset, output_path)
    elif output_path.suffix == ".zarr":
        delayed = _prepare_download_dataset_as_zarr(dataset, output_path)
    else:
        delayed = _prepare_download_dataset_as_netcdf(dataset, output_path)
    return delayed


def download_delayed_dataset(
    delayed: Delayed, disable_progress_bar: bool
) -> None:
    if disable_progress_bar:
        delayed.compute()
    else:
        with TqdmCallback():
            delayed.compute()


def _prepare_download_dataset_as_netcdf(
    dataset: xarray.Dataset, output_path: pathlib.Path
):
    logger.debug("Writing dataset to NetCDF")
    return dataset.to_netcdf(output_path, mode="w", compute=False)


def _prepare_download_dataset_as_zarr(
    dataset: xarray.Dataset, output_path: pathlib.Path
):
    logger.debug("Writing dataset to Zarr")
    store = zarr.DirectoryStore(output_path)
    return dataset.to_zarr(store=store, mode="w", compute=False)



import pathlib
from typing import List, Optional, Union

from copernicus_marine_client.core_functions.get import get_function
from copernicus_marine_client.python_interface.exception_handler import (
    log_exception_and_exit,
)


@log_exception_and_exit
def get(
    dataset_url: Optional[str] = None,
    dataset_id: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    no_directories: bool = False,
    show_outputnames: bool = False,
    output_directory: Optional[Union[pathlib.Path, str]] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    force_download: bool = False,
    overwrite_output_data: bool = False,
    request_file: Optional[Union[pathlib.Path, str]] = None,
    force_service: Optional[str] = None,
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    filter: Optional[str] = None,
    regex: Optional[str] = None,
    disable_progress_bar: bool = False,
    staging: bool = False,
) -> List[pathlib.Path]:
    """
    Fetches data from the Copernicus Marine server based on the provided parameters.

    Args:
        dataset_url (str, optional): The URL of the dataset to retrieve.
        dataset_id (str, optional): The unique identifier of the dataset.
        username (str, optional): The username for authentication.
        password (str, optional): The password for authentication.
        no_directories (bool, optional): If True, downloaded files will not be organized into directories.
        show_outputnames (bool, optional): If True, display the names of the downloaded files.
        output_directory (Union[pathlib.Path, str], optional): The directory where downloaded files will be saved.
        credentials_file (Union[pathlib.Path, str], optional): Path to a file containing authentication credentials.
        force_download (bool, optional): Skip confirmation before download.
        overwrite_output_data (bool, optional): If True, overwrite existing output files.
        request_file (Union[pathlib.Path, str], optional): Path to a file containing request parameters.
        force_service (str, optional): Force the use of a specific service.
        overwrite_metadata_cache (bool, optional): If True, overwrite the metadata cache.
        no_metadata_cache (bool, optional): If True, do not use the metadata cache.
        filter (str, optional): Apply a filter to the downloaded data.
        regex (str, optional): Apply a regular expression filter to the downloaded data.

    Returns:
        List[pathlib.Path]: A list of paths to the downloaded files.
    """  # noqa
    output_directory = (
        pathlib.Path(output_directory) if output_directory else None
    )
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )
    request_file = pathlib.Path(request_file) if request_file else None
    return get_function(
        dataset_url,
        dataset_id,
        username,
        password,
        no_directories,
        show_outputnames,
        output_directory,
        credentials_file,
        force_download,
        overwrite_output_data,
        request_file,
        force_service,
        overwrite_metadata_cache,
        no_metadata_cache,
        filter,
        regex,
        disable_progress_bar,
        staging=staging,
    )



import pathlib
from typing import Optional

from copernicus_marine_client.core_functions.login import login_function
from copernicus_marine_client.core_functions.utils import (
    DEFAULT_CLIENT_BASE_DIRECTORY,
)


def login(
    username: Optional[str] = None,
    password: Optional[str] = None,
    configuration_file_directory: pathlib.Path = DEFAULT_CLIENT_BASE_DIRECTORY,
    overwrite_configuration_file: bool = False,
) -> bool:
    return login_function(
        username=username,
        password=password,
        configuration_file_directory=configuration_file_directory,
        overwrite_configuration_file=overwrite_configuration_file,
    )



from functools import wraps
from typing import Callable

import click


def log_exception_and_exit(function: Callable):
    @wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except click.Abort:
            print("Abort")
        except Exception as exception:
            raise exception

    return wrapper



import asyncio
import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from importlib.metadata import version as package_version
from itertools import groupby, repeat
from json import loads
from typing import Any, Dict, Iterator, List, Optional, Tuple, TypeVar, Union

import aiohttp
import nest_asyncio
import pystac
import requests
from cachier import cachier
from tqdm import tqdm

from copernicus_marine_client.command_line_interface.exception_handler import (
    log_exception_debug,
)
from copernicus_marine_client.core_functions.utils import (
    DEFAULT_CLIENT_BASE_DIRECTORY,
    construct_query_params_for_marine_data_store_monitoring,
    map_reject_none,
    next_or_raise_exception,
)

logger = logging.getLogger("copernicus_marine_root_logger")

_S = TypeVar("_S")
_T = TypeVar("_T")


class _ServiceName(str, Enum):
    MOTU = "motu"
    OPENDAP = "opendap"
    GEOSERIES = "arco-geo-series"
    TIMESERIES = "arco-time-series"
    FILES = "original-files"
    FTP = "ftp"
    WMS = "wms"
    WMTS = "wmts"


class _ServiceShortName(str, Enum):
    MOTU = "motu"
    OPENDAP = "opendap"
    GEOSERIES = "geoseries"
    TIMESERIES = "timeseries"
    FILES = "files"
    FTP = "ftp"
    WMS = "wms"
    WMTS = "wmts"


MARINE_DATA_STORE_STAC_BASE_URL = "https://stac.marine.copernicus.eu/metadata"
MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL = (
    MARINE_DATA_STORE_STAC_BASE_URL + "/catalog.stac.json"
)
MARINE_DATA_STORE_STAC_BASE_URL_STAGING = (
    "https://stac-dta.marine.copernicus.eu/metadata"
)
MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL_STAGING = (
    MARINE_DATA_STORE_STAC_BASE_URL_STAGING + "/catalog.stac.json"
)


@dataclass(frozen=True)
class _Service:
    service_name: _ServiceName
    short_name: _ServiceShortName

    def aliases(self) -> List[str]:
        return (
            [self.service_name.value, self.short_name.value]
            if self.short_name.value != self.service_name.value
            else [self.service_name.value]
        )

    def to_json_dict(self):
        return {
            "service_name": self.service_name.value,
            "short_name": self.short_name.value,
        }


class CopernicusMarineDatasetServiceType(_Service, Enum):
    MOTU = _ServiceName.MOTU, _ServiceShortName.MOTU
    OPENDAP = _ServiceName.OPENDAP, _ServiceShortName.OPENDAP
    GEOSERIES = _ServiceName.GEOSERIES, _ServiceShortName.GEOSERIES
    TIMESERIES = (
        _ServiceName.TIMESERIES,
        _ServiceShortName.TIMESERIES,
    )
    FILES = _ServiceName.FILES, _ServiceShortName.FILES
    FTP = _ServiceName.FTP, _ServiceShortName.FTP
    WMS = _ServiceName.WMS, _ServiceShortName.WMS
    WMTS = _ServiceName.WMS, _ServiceShortName.WMTS


def _service_type_from_web_api_string(
    name: str,
) -> CopernicusMarineDatasetServiceType:
    class WebApi(Enum):
        MOTU = "motu"
        OPENDAP = "opendap"
        GEOSERIES = "timeChunked"
        TIMESERIES = "geoChunked"
        FILES = "native"
        FTP = "ftp"
        WMS = "wms"
        WMTS = "wmts"

    web_api_mapping = {
        WebApi.MOTU: CopernicusMarineDatasetServiceType.MOTU,
        WebApi.OPENDAP: CopernicusMarineDatasetServiceType.OPENDAP,
        WebApi.GEOSERIES: CopernicusMarineDatasetServiceType.GEOSERIES,
        WebApi.TIMESERIES: CopernicusMarineDatasetServiceType.TIMESERIES,
        WebApi.FILES: CopernicusMarineDatasetServiceType.FILES,
        WebApi.FTP: CopernicusMarineDatasetServiceType.FTP,
        WebApi.WMS: CopernicusMarineDatasetServiceType.WMS,
        WebApi.WMTS: CopernicusMarineDatasetServiceType.WMTS,
    }

    return next_or_raise_exception(
        (
            service_type
            for service_web_api, service_type in web_api_mapping.items()
            if service_web_api.value == name
        ),
        ServiceNotHandled(name),
    )


class ServiceNotHandled(Exception):
    ...


VERSION_DEFAULT = "default"
PART_DEFAULT = "default"


@dataclass
class CopernicusMarineCoordinates:
    coordinates_id: str
    units: str
    minimum_value: Optional[float]
    maximum_value: Optional[float]
    step: Optional[float]
    values: Optional[str]


@dataclass
class CopernicusMarineVariable:
    short_name: str
    standard_name: str
    units: str
    bbox: Tuple[float, float, float, float]
    coordinates: list[CopernicusMarineCoordinates]


@dataclass
class CopernicusMarineService:
    service_type: CopernicusMarineDatasetServiceType
    uri: str
    variables: list[CopernicusMarineVariable]


@dataclass
class CopernicusMarineVersionPart:
    name: str
    services: list[CopernicusMarineService]

    def get_service_by_service_type(
        self, service_type: CopernicusMarineDatasetServiceType
    ):
        return next(
            service
            for service in self.services
            if service.service_type == service_type
        )


class DatasetPartNotFound(Exception):
    ...


@dataclass
class CopernicusMarineDatasetVersion:
    label: str
    parts: list[CopernicusMarineVersionPart]

    def get_part_by_name(self, name: str) -> CopernicusMarineVersionPart:
        for part in self.parts:
            if part.name == name:
                return part
        raise DatasetPartNotFound(f"Part {name} not found")


class DatasetVersionNotFound(Exception):
    ...


@dataclass
class CopernicusMarineProductDataset:
    dataset_id: str
    dataset_name: str
    versions: list[CopernicusMarineDatasetVersion]

    def _seperate_version_and_default(
        self,
    ) -> Tuple[
        Optional[CopernicusMarineDatasetVersion],
        List[CopernicusMarineDatasetVersion],
    ]:
        default_version = None
        versions = []
        for version in self.versions:
            if version.label == VERSION_DEFAULT:
                default_version = version
            else:
                versions.append(version)
        return default_version, versions

    def get_latest_version(
        self, raise_not_found: bool = True
    ) -> Optional[CopernicusMarineDatasetVersion]:
        default_version, versions = self._seperate_version_and_default()
        sorted_versions = sorted(versions, key=lambda x: x.label)
        if sorted_versions:
            return sorted_versions[-1]
        else:
            if default_version:
                return default_version
            else:
                if raise_not_found:
                    raise DatasetVersionNotFound(
                        f"No version found for dataset {self.dataset_id}"
                    )
                else:
                    return None


@dataclass
class CopernicusMarineProductProvider:
    name: str
    roles: list[str]
    url: str
    email: str


@dataclass
class CopernicusMarineProduct:
    title: str
    product_id: str
    thumbnail_url: str
    description: str
    digital_object_identifier: Optional[str]
    sources: List[str]
    processing_level: Optional[str]
    production_center: str
    keywords: dict[str, str]
    datasets: list[CopernicusMarineProductDataset]


@dataclass
class ProductDatasetParser(ABC):
    dataset_id: str
    dataset_name: str
    versions: list[CopernicusMarineDatasetVersion]

    @abstractmethod
    def to_copernicus_marine_dataset(self) -> CopernicusMarineProductDataset:
        ...


@dataclass
class ProductParser(ABC):
    title: str
    product_id: str
    thumbnail_url: str
    description: str
    digital_object_identifier: Optional[str]
    sources: List[str]
    processing_level: Optional[str]
    production_center: str
    keywords: dict[str, str]


@dataclass
class ProductDatasetFromPortal(ProductDatasetParser):
    raw_services: Dict
    layer_elements: List

    def to_copernicus_marine_dataset(self) -> CopernicusMarineProductDataset:
        copernicus_marine_dataset = CopernicusMarineProductDataset(
            dataset_id=self.dataset_id,
            dataset_name=self.dataset_name,
            versions=self.versions,
        )
        latest_version = copernicus_marine_dataset.get_latest_version(
            raise_not_found=False
        )

        portal_services = portal_services_to_services(
            self.raw_services,
            self.layer_elements,
        )

        if latest_version:
            latest_version_part = latest_version.parts[0]
            latest_version_part.services += portal_services
        elif portal_services:
            default_version = CopernicusMarineDatasetVersion(
                label=VERSION_DEFAULT,
                parts=[
                    CopernicusMarineVersionPart(
                        name=PART_DEFAULT, services=portal_services
                    )
                ],
            )
            copernicus_marine_dataset.versions = [default_version]

        return copernicus_marine_dataset


@dataclass
class ProductFromPortal(ProductParser):
    datasets: list[ProductDatasetFromPortal]

    def to_copernicus_marine_product(self) -> CopernicusMarineProduct:
        return CopernicusMarineProduct(
            title=self.title,
            product_id=self.product_id,
            thumbnail_url=self.thumbnail_url,
            description=self.description,
            digital_object_identifier=self.digital_object_identifier,
            sources=self.sources,
            processing_level=self.processing_level,
            production_center=self.production_center,
            keywords=self.keywords,
            datasets=[
                dataset.to_copernicus_marine_dataset()
                for dataset in self.datasets
            ],
        )


@dataclass
class ProductDatasetFromMarineDataStore(ProductDatasetParser):
    def to_copernicus_marine_dataset(self) -> CopernicusMarineProductDataset:
        return CopernicusMarineProductDataset(
            dataset_id=self.dataset_id,
            dataset_name=self.dataset_name,
            versions=self.versions,
        )


@dataclass
class ProductFromMarineDataStore(ProductParser):
    datasets: list[ProductDatasetFromMarineDataStore]

    def to_copernicus_marine_product(self) -> CopernicusMarineProduct:
        return CopernicusMarineProduct(
            title=self.title,
            product_id=self.product_id,
            thumbnail_url=self.thumbnail_url,
            description=self.description,
            digital_object_identifier=self.digital_object_identifier,
            sources=self.sources,
            processing_level=self.processing_level,
            production_center=self.production_center,
            keywords=self.keywords,
            datasets=[
                dataset.to_copernicus_marine_dataset()
                for dataset in self.datasets
            ],
        )


@dataclass
class CopernicusMarineCatalogue:
    products: list[CopernicusMarineProduct]

    def filter(self, tokens: list[str]):
        return filter_catalogue_with_strings(self, tokens)


class CatalogParserConnection:
    def __init__(self, proxy: Optional[str] = None) -> None:
        self.proxy = proxy
        self.session = aiohttp.ClientSession(trust_env=True)

    async def get_json_file(self, url: str) -> dict[str, Any]:
        logger.debug(f"Fetching json file at this url: {url}")
        async with self.session.get(
            url,
            params=construct_query_params_for_marine_data_store_monitoring(),
        ) as response:
            return await response.json()

    async def close(self) -> None:
        await self.session.close()


def _construct_copernicus_marine_service(
    stac_service_name, stac_asset, datacube
) -> Optional[CopernicusMarineService]:
    try:
        service_uri = stac_asset.get_absolute_href()
        if not service_uri.endswith("/"):
            return CopernicusMarineService(
                service_type=_service_type_from_web_api_string(
                    stac_service_name
                ),
                uri=stac_asset.get_absolute_href(),
                variables=_get_variables(datacube),
            )
        else:
            return None
    except ServiceNotHandled as service_not_handled:
        log_exception_debug(service_not_handled)
        return None


def _get_versions_from_marine_datastore(
    datacubes: List[pystac.Item],
) -> List[CopernicusMarineDatasetVersion]:
    copernicus_marine_dataset_versions: List[
        CopernicusMarineDatasetVersion
    ] = []
    for datacube in datacubes:
        _, dataset_version, _ = get_version_and_part_from_full_dataset_id(
            datacube.id
        )
        if dataset_version not in map(
            lambda x: x.label, copernicus_marine_dataset_versions
        ):
            parts = _get_parts(datacube)
            if parts:
                copernicus_marine_dataset_versions.append(
                    CopernicusMarineDatasetVersion(
                        label=dataset_version, parts=parts
                    )
                )
    return copernicus_marine_dataset_versions


def _get_parts(datacube: pystac.Item) -> List[CopernicusMarineVersionPart]:
    services = _get_services(datacube)
    if services:
        return [
            CopernicusMarineVersionPart(
                name=PART_DEFAULT, services=_get_services(datacube)
            )
        ]
    else:
        return []


def _get_services(
    datacube: pystac.Item,
) -> list[CopernicusMarineService]:
    stac_assets_dict = datacube.get_assets()
    return [
        dataset_service
        for stac_service_name, stac_asset in stac_assets_dict.items()
        if stac_asset.roles and "data" in stac_asset.roles
        if (
            dataset_service := _construct_copernicus_marine_service(
                stac_service_name, stac_asset, datacube
            )
        )
        is not None
    ]


def _get_coordinates(
    dimensions_cube: dict,
) -> dict[str, CopernicusMarineCoordinates]:
    def _create_coordinate(
        key: str, value: dict
    ) -> CopernicusMarineCoordinates:
        return CopernicusMarineCoordinates(
            coordinates_id="depth" if key == "elevation" else key,
            units=value.get("unit") or "",
            minimum_value=value["extent"][0] if "extent" in value else None,
            maximum_value=value["extent"][1] if "extent" in value else None,
            step=value.get("step"),
            values=value.get("values"),
        )

    coordinates_dict = {}
    for key, value in dimensions_cube.items():
        coordinates_dict[key] = _create_coordinate(key, value)
    return coordinates_dict


def _get_variables(
    stac_dataset: pystac.Item,
) -> list[CopernicusMarineVariable]:
    def _create_variable(
        variable_cube: dict[str, Any],
        bbox: tuple[float, float, float, float],
        coordinates_dict: dict[str, CopernicusMarineCoordinates],
    ) -> Union[CopernicusMarineVariable, None]:
        coordinates = variable_cube["dimensions"]
        return CopernicusMarineVariable(
            short_name=variable_cube["id"],
            standard_name=variable_cube["standardName"],
            units=variable_cube.get("unit") or "",
            bbox=bbox,
            coordinates=[coordinates_dict[key] for key in coordinates],
        )

    coordinates_dict = _get_coordinates(
        stac_dataset.properties["cube:dimensions"]
    )
    bbox = stac_dataset.bbox
    variables: list[Optional[CopernicusMarineVariable]] = []
    for var_cube in stac_dataset.properties["cube:variables"].values():
        variables += [_create_variable(var_cube, bbox, coordinates_dict)]
    return [var for var in variables if var]


def _construct_marine_data_store_dataset(
    datacubes_by_id: List,
) -> Optional[ProductDatasetFromMarineDataStore]:
    dataset_id = datacubes_by_id[0]
    datacubes = list(datacubes_by_id[1])
    if datacubes:
        versions = _get_versions_from_marine_datastore(datacubes)
        return ProductDatasetFromMarineDataStore(
            dataset_id=dataset_id,
            dataset_name=datacubes[0].properties["title"],
            versions=versions,
        )
    return None


def _construct_marine_data_store_product(
    stac_tuple: Tuple[pystac.Collection, List[pystac.Item]],
) -> ProductFromMarineDataStore:
    stac_product, stac_datasets = stac_tuple
    stac_datasets_sorted = sorted(stac_datasets, key=lambda x: x.id)
    datacubes_by_id = groupby(
        stac_datasets_sorted,
        key=lambda x: get_version_and_part_from_full_dataset_id(x.id)[0],
    )

    datasets = map(
        _construct_marine_data_store_dataset, datacubes_by_id  # type: ignore
    )

    production_center = [
        provider.name
        for provider in stac_product.providers or []
        if "producer" in provider.roles
    ]

    production_center_name = production_center[0] if production_center else ""

    thumbnail = stac_product.assets and stac_product.assets.get("thumbnail")
    digital_object_identifier = (
        stac_product.extra_fields.get("sci:doi", None)
        if stac_product.extra_fields
        else None
    )
    sources = _get_stac_product_property(stac_product, "sources") or []
    processing_level = _get_stac_product_property(
        stac_product, "processingLevel"
    )

    return ProductFromMarineDataStore(
        title=stac_product.title or stac_product.id,
        product_id=stac_product.id,
        thumbnail_url=thumbnail.get_absolute_href() if thumbnail else "",
        description=stac_product.description,
        digital_object_identifier=digital_object_identifier,
        sources=sources,
        processing_level=processing_level,
        production_center=production_center_name,
        keywords=stac_product.keywords,
        datasets=sorted(
            [dataset for dataset in datasets if dataset],
            key=lambda dataset: dataset.dataset_id,
        ),
    )


def _get_stac_product_property(
    stac_product: pystac.Collection, property_key: str
) -> Optional[Any]:
    properties: Dict[str, str] = (
        stac_product.extra_fields.get("properties", {})
        if stac_product.extra_fields
        else {}
    )
    return properties.get(property_key)


async def async_fetch_items_from_collection(
    root_url: str,
    connection: CatalogParserConnection,
    collection: pystac.Collection,
) -> List[pystac.Item]:
    items = []
    for link in collection.get_item_links():
        if not link.owner:
            logger.warning(f"Invalid Item, no owner for: {link.href}")
            continue
        url = root_url + "/" + link.owner.id + "/" + link.href
        try:
            item_json = await connection.get_json_file(url)
            items.append(pystac.Item.from_dict(item_json))
        except pystac.STACError as exception:
            message = (
                "Invalid Item: If datetime is None, a start_datetime "
                + "and end_datetime must be supplied."
            )
            if exception.args[0] != message:
                logger.error(exception)
                raise pystac.STACError(exception.args)
    return items


async def async_fetch_collection(
    root_url: str, connection: CatalogParserConnection, url: str
) -> Optional[Tuple[pystac.Collection, List[pystac.Item]]]:
    json_collection = await connection.get_json_file(url)
    try:
        collection = pystac.Collection.from_dict(json_collection)
        items = await async_fetch_items_from_collection(
            root_url, connection, collection
        )
        return (collection, items)

    except KeyError as exception:
        messages = ["spatial", "temporal"]
        if exception.args[0] not in messages:
            logger.error(exception)
        return None


async def async_fetch_childs(
    root_url: str,
    connection: CatalogParserConnection,
    child_links: List[pystac.Link],
) -> Iterator[Optional[Tuple[pystac.Collection, List[pystac.Item]]]]:
    tasks = []
    for link in child_links:
        tasks.append(
            asyncio.ensure_future(
                async_fetch_collection(
                    root_url, connection, link.absolute_href
                )
            )
        )
    return filter(lambda x: x is not None, await asyncio.gather(*tasks))


async def async_fetch_catalog(
    connection: CatalogParserConnection,
    staging: bool = False,
) -> Tuple[pystac.Catalog, Iterator[pystac.Collection]]:
    catalog_root_url = (
        MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL
        if not staging
        else MARINE_DATA_STORE_STAC_ROOT_CATALOG_URL_STAGING
    )
    json_catalog = await connection.get_json_file(catalog_root_url)
    catalog = pystac.Catalog.from_dict(json_catalog)
    child_links = catalog.get_child_links()
    root_url = (
        MARINE_DATA_STORE_STAC_BASE_URL
        if not staging
        else (MARINE_DATA_STORE_STAC_BASE_URL_STAGING)
    )
    childs = await async_fetch_childs(root_url, connection, child_links)
    return catalog, childs


def _retrieve_marine_data_store_products(
    connection: CatalogParserConnection,
    staging: bool = False,
) -> list[ProductFromMarineDataStore]:
    nest_asyncio.apply()
    loop = asyncio.get_event_loop()
    _, marine_data_store_root_collections = loop.run_until_complete(
        async_fetch_catalog(connection=connection, staging=staging)
    )

    products = map_reject_none(
        _construct_marine_data_store_product,
        marine_data_store_root_collections,
    )

    return list(products)


def parse_catalogue(
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    disable_progress_bar: bool,
    staging: bool = False,
) -> CopernicusMarineCatalogue:
    logger.debug("Parsing catalogue...")
    catalog = _parse_catalogue(
        overwrite_cache=overwrite_metadata_cache,
        ignore_cache=no_metadata_cache,
        _version=package_version("copernicus-marine-client"),
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )
    logger.debug("Catalogue parsed")
    return catalog


def merge_products(
    products_from_marine_data_store: List[ProductFromMarineDataStore],
    products_from_portal: List[ProductFromPortal],
) -> List[CopernicusMarineProduct]:
    merged_products: List[CopernicusMarineProduct] = []
    for marine_data_store_product in products_from_marine_data_store:
        merged_products.append(
            marine_data_store_product.to_copernicus_marine_product()
        )

    for portal_product in products_from_portal:
        maybe_merged_product = list(
            filter(
                lambda x: x.product_id == portal_product.product_id,
                merged_products,
            )
        )
        if not maybe_merged_product:
            merged_product = portal_product.to_copernicus_marine_product()
            merged_products.append(merged_product)

        else:
            merged_product = maybe_merged_product[0]
            for portal_dataset in portal_product.datasets:
                maybe_merged_dataset = list(
                    filter(
                        lambda x: x.dataset_id == portal_dataset.dataset_id,
                        merged_product.datasets,
                    )
                )
                if not maybe_merged_dataset:
                    merged_product.datasets.append(
                        portal_dataset.to_copernicus_marine_dataset()
                    )
                else:
                    merged_dataset = maybe_merged_dataset[0]
                    for portal_version in portal_dataset.versions:
                        maybe_merged_version = list(
                            filter(
                                lambda x: x.label == portal_version.label,
                                merged_dataset.versions,
                            )
                        )
                        if not maybe_merged_version:
                            merged_dataset.versions.append(portal_version)
                        else:
                            merged_version = maybe_merged_version[0]
                            for portal_part in portal_version.parts:
                                maybe_merged_part = list(
                                    filter(
                                        lambda x: x.name == portal_part.name,
                                        merged_version.parts,
                                    )
                                )
                                if not maybe_merged_part:
                                    merged_version.parts.append(portal_part)
                                else:
                                    merged_part = maybe_merged_part[0]
                                    for portal_service in portal_part.services:
                                        maybe_merged_service = list(
                                            filter(
                                                lambda x: x.service_type
                                                == portal_service.service_type,
                                                merged_part.services,
                                            )
                                        )
                                        if not maybe_merged_service:
                                            merged_part.services.append(
                                                portal_service
                                            )
                                        else:
                                            continue
                                    merged_part.services.sort(
                                        key=lambda x: x.service_type.service_name.value
                                    )
                            merged_version.parts.sort(key=lambda x: x.name)
                    merged_dataset.versions.sort(
                        key=lambda x: x.label
                        if x.label != VERSION_DEFAULT
                        else "110001"
                    )
                    latest_version = merged_dataset.get_latest_version(
                        raise_not_found=False
                    )
                    if latest_version:
                        latest_version_part = latest_version.parts[0]
                        latest_version_part.services += (
                            portal_services_to_services(
                                portal_dataset.raw_services,
                                portal_dataset.layer_elements,
                            )
                        )

            merged_product.datasets.sort(key=lambda x: x.dataset_id)

    return merged_products


@cachier(cache_dir=DEFAULT_CLIENT_BASE_DIRECTORY)
def _parse_catalogue(
    _version: str,  # force cachier to overwrite cache in case of version update
    disable_progress_bar: bool,
    staging: bool = False,
) -> CopernicusMarineCatalogue:
    progress_bar = tqdm(
        total=4, desc="Fetching catalog", disable=disable_progress_bar
    )
    connection = CatalogParserConnection()
    if not staging:
        logger.debug("Parsing portal catalogue...")
        portal_products = _parse_portal_backend_products(connection)
        progress_bar.update()
        logger.debug("Portal catalogue parsed")
    else:
        # because of data misalignment we don't want to use data-be endpoint
        portal_products = []

    marine_data_store_products = _retrieve_marine_data_store_products(
        connection=connection, staging=staging
    )
    progress_bar.update()

    products_merged = merge_products(
        marine_data_store_products, portal_products
    )
    products_merged.sort(key=lambda x: x.product_id)
    progress_bar.update()

    full_catalog = CopernicusMarineCatalogue(products=products_merged)

    progress_bar.update()
    asyncio.run(connection.close())

    return full_catalog


async def _async_fetch_raw_products(
    product_ids: List[str], connection: CatalogParserConnection
):
    tasks = []
    for product_id in product_ids:
        tasks.append(
            asyncio.ensure_future(
                connection.get_json_file(product_url(product_id))
            )
        )

    return await asyncio.gather(*tasks)


def product_url(product_id: str) -> str:
    base_url = "https://data-be-prd.marine.copernicus.eu/api/dataset"
    return f"{base_url}/{product_id}" + "?variant=detailed-v2"


def variable_title_to_standard_name(variable_title: str) -> str:
    return variable_title.lower().replace(" ", "_")


def variable_to_pick(layer: dict[str, Any]) -> bool:
    return (
        layer["variableId"] != "__DEFAULT__"
        and layer["subsetVariableIds"]
        and len(layer["subsetVariableIds"]) == 1
    )


def _to_service(
    service_name: str, service_url: str, layer_elements
) -> Optional[CopernicusMarineService]:
    """
    Simon question for review: variables are parsed from the 'layer' field and this
    function is used by both "mds_stac_to_services" and "portal_to_services" functions.
    Which means that we are adding the same variables for all services.
    Is it what we want to begin?
    """
    try:
        service_type = _service_type_from_web_api_string(service_name)
        if not service_url.endswith("thredds/dodsC/"):
            return CopernicusMarineService(
                service_type=service_type,
                uri=service_url,
                variables=list(
                    map(to_variable, filter(variable_to_pick, layer_elements))
                ),
            )
        else:
            return None
    except ServiceNotHandled as service_not_handled:
        log_exception_debug(service_not_handled)
        return None


def to_coordinates(
    subset_attributes: Tuple[str, dict[str, Any]], layer: dict[str, Any]
) -> CopernicusMarineCoordinates:
    coordinate_name = subset_attributes[0]
    values: Optional[str]
    if coordinate_name == "depth":
        values = layer.get("zValues")
    elif coordinate_name == "time":
        values = layer.get("tValues")
    else:
        values = None
    return CopernicusMarineCoordinates(
        coordinates_id=subset_attributes[0],
        units=subset_attributes[1]["units"],
        minimum_value=subset_attributes[1]["min"],
        maximum_value=subset_attributes[1]["max"],
        step=subset_attributes[1].get("step"),
        values=values,
    )


def to_variable(layer: dict[str, Any]) -> CopernicusMarineVariable:
    return CopernicusMarineVariable(
        short_name=layer["variableId"],
        standard_name=variable_title_to_standard_name(layer["variableTitle"]),
        units=layer["units"],
        bbox=layer["bbox"],
        coordinates=list(
            map(to_coordinates, layer["subsetAttrs"].items(), repeat(layer))
        ),
    )


@dataclass
class DistinctDatasetVersion:
    dataset_id: str
    dataset_version: str
    layer_elements: List
    raw_services: Dict
    stac_items_values: Optional[Dict]


def mds_stac_to_services(
    distinct_dataset_version: DistinctDatasetVersion,
) -> List[CopernicusMarineService]:
    copernicus_marine_services = []

    stac_assets = (
        distinct_dataset_version.stac_items_values
        and distinct_dataset_version.stac_items_values.get("assets", None)
    )
    if stac_assets:
        for (
            service_name,
            service_url,
        ) in stac_assets.items():
            service = _to_service(
                service_name,
                service_url["href"],
                distinct_dataset_version.layer_elements,
            )
            if service:
                copernicus_marine_services.append(service)

    return copernicus_marine_services


def portal_services_to_services(
    raw_services: Dict, layer_elements: List
) -> list[CopernicusMarineService]:
    copernicus_marine_services = []

    for (
        service_name,
        service_url,
    ) in raw_services.items():
        copernicus_marine_services.append(
            _to_service(
                service_name,
                service_url,
                layer_elements,
            )
        )

    return [service for service in copernicus_marine_services if service]


def to_dataset(
    distinct_dataset_versions: List[DistinctDatasetVersion],
) -> Optional[ProductDatasetFromPortal]:
    if distinct_dataset_versions:
        first_distinct_dataset_version = distinct_dataset_versions[0]
        dataset_id = first_distinct_dataset_version.dataset_id
        layer_elements = list(first_distinct_dataset_version.layer_elements)
        sub_dataset_title = (
            layer_elements[0]["subdatasetTitle"]
            if layer_elements
            else dataset_id
        )
        return ProductDatasetFromPortal(
            dataset_id=dataset_id,
            dataset_name=sub_dataset_title,
            versions=list(
                map_reject_none(to_version, distinct_dataset_versions)
            ),
            raw_services=first_distinct_dataset_version.raw_services,
            layer_elements=first_distinct_dataset_version.layer_elements,
        )
    return None


def to_part(
    distinct_dataset_version: DistinctDatasetVersion,
) -> List[CopernicusMarineVersionPart]:
    # backend_portal_services = portal_to_services(distinct_dataset_version)
    mds_stac_services = mds_stac_to_services(distinct_dataset_version)
    if mds_stac_services:
        return [
            CopernicusMarineVersionPart(
                name=PART_DEFAULT,  # Because parts not treated yet
                services=mds_stac_services,
            )
        ]
    else:
        return []


def to_version(
    distinct_dataset_version: DistinctDatasetVersion,
) -> Optional[CopernicusMarineDatasetVersion]:
    parts = to_part(distinct_dataset_version)
    if parts:
        return CopernicusMarineDatasetVersion(
            label=distinct_dataset_version.dataset_version,
            parts=parts,
        )
    else:
        return None


REGEX_PATTERN_DATE_YYYYMM = r"[12]\d{3}(0[1-9]|1[0-2])"


def get_version_and_part_from_full_dataset_id(
    full_dataset_id: str,
) -> Tuple[str, str, str]:
    if "--ext--" in full_dataset_id:
        name_with_maybe_version, _ = full_dataset_id.split("--ext--")
    else:
        name_with_maybe_version = full_dataset_id
        # part = PART_DEFAULT
    pattern = rf"^(.*?)(?:_({REGEX_PATTERN_DATE_YYYYMM}))?$"
    match = re.match(pattern, name_with_maybe_version)
    if match:
        dataset_name = match.group(1)
        version = match.group(2) or VERSION_DEFAULT
    else:
        raise Exception(f"Could not parse dataset id: {full_dataset_id}")
    return dataset_name, version, PART_DEFAULT


def construct_unique_dataset(
    raw_service, groups_layers, stac_items
) -> List[DistinctDatasetVersion]:
    dataset_id_from_raw_service = raw_service[0]
    groups_layer = [
        group_layer
        for group_layer in groups_layers.items()
        if group_layer[0] == dataset_id_from_raw_service
    ]
    dataset_layer_elements = groups_layer[0][1] if groups_layer else []
    dataset_raw_services = raw_service[1]

    dataset_versions = []

    for stac_dataset_id, stac_items_values in stac_items.items():
        if stac_dataset_id.startswith(dataset_id_from_raw_service):
            if (
                "--ext--"
                in stac_dataset_id.split(dataset_id_from_raw_service)[-1]
            ):
                continue
            else:
                (
                    _,
                    dataset_version,
                    _,
                ) = get_version_and_part_from_full_dataset_id(stac_dataset_id)

                # Simon question for review: Here, it has been decided, for each
                # version of a dataset, to fill all services (raw + stac) with the
                # variables from dataset_layer_elements. Is it OK ?
                dataset_versions.append(
                    DistinctDatasetVersion(
                        dataset_id=dataset_id_from_raw_service,
                        dataset_version=dataset_version,
                        layer_elements=dataset_layer_elements,
                        raw_services=dataset_raw_services,
                        stac_items_values=stac_items_values,
                    )
                )
    else:
        _, dataset_version, _ = get_version_and_part_from_full_dataset_id(
            dataset_id_from_raw_service
        )
        dataset_versions.append(
            DistinctDatasetVersion(
                dataset_id=dataset_id_from_raw_service,
                dataset_version=dataset_version,
                layer_elements=dataset_layer_elements,
                raw_services=dataset_raw_services,
                stac_items_values=None,
            )
        )

    return dataset_versions


def to_datasets(
    raw_services: dict[str, dict[str, str]],
    layers: dict[str, dict[str, Any]],
    stac_items: dict,
) -> list[ProductDatasetFromPortal]:
    groups_layers = defaultdict(list)
    for layer in layers.values():
        subdataset_id = layer["subdatasetId"]
        groups_layers[subdataset_id].append(layer)

    distinct_dataset_versions = map(
        construct_unique_dataset,
        raw_services.items(),
        repeat(groups_layers),
        repeat(stac_items),
    )

    return sorted(
        map_reject_none(to_dataset, distinct_dataset_versions),
        key=lambda distinct_dataset: distinct_dataset.dataset_id,
    )


def _parse_portal_product(raw_product: dict[str, Any]) -> ProductFromPortal:
    return ProductFromPortal(
        title=raw_product["title"],
        product_id=raw_product["id"],
        thumbnail_url=raw_product["thumbnailUrl"],
        description=raw_product["abstract"],
        digital_object_identifier=raw_product["doi"],
        sources=raw_product["sources"],
        processing_level=raw_product["processingLevel"]
        if "processingLevel" in raw_product
        else None,
        production_center=raw_product["originatingCenter"],
        keywords=raw_product["keywords"],
        datasets=to_datasets(
            raw_product["services"],
            raw_product["layers"],
            raw_product["stacItems"],
        ),
    )


def _parse_portal_backend_products(
    connection: CatalogParserConnection,
) -> List[ProductFromPortal]:
    base_url = "https://data-be-prd.marine.copernicus.eu/api/datasets"
    response = requests.post(
        base_url,
        json={"size": 1000, "includeOmis": True},
    )
    assert response.ok, response.text
    raw_catalogue: dict[str, Any] = loads(response.text)

    nest_asyncio.apply()
    loop = asyncio.get_event_loop()

    results = loop.run_until_complete(
        _async_fetch_raw_products(raw_catalogue["datasets"].keys(), connection)
    )
    return list(map(_parse_portal_product, results))


# ---------------------------------------
# --- Utils function on any catalogue ---
# ---------------------------------------


def get_dataset_and_version_part_from_id(
    catalogue: CopernicusMarineCatalogue,
    dataset_id: str,
    part_name: str = PART_DEFAULT,
) -> Tuple[CopernicusMarineProductDataset, CopernicusMarineVersionPart]:
    for product in catalogue.products:
        for dataset in product.datasets:
            if dataset_id == dataset.dataset_id:
                dataset_version = dataset.get_latest_version(
                    raise_not_found=True
                )
                dataset_part = dataset_version.get_part_by_name(  # type: ignore
                    part_name
                )
                return dataset, dataset_part
    error = KeyError(
        f"The requested dataset '{dataset_id}' was not found in the catalogue,"
        " you can use 'copernicus-marine describe --include-datasets "
        "--contains <search_token>' to find datasets"
    )
    raise error


def get_dataset_and_suffix_path_from_url(
    catalogue: CopernicusMarineCatalogue, dataset_url: str
) -> Tuple[CopernicusMarineProductDataset, str]:
    return next_or_raise_exception(
        (
            dataset_and_suffix_url
            for product in catalogue.products
            for dataset in product.datasets
            for service in dataset.get_latest_version(  # type: ignore
                raise_not_found=True
            )
            .get_part_by_name(PART_DEFAULT)
            .services
            if (
                dataset_and_suffix_url := _parse_dataset_and_suffix_url(
                    dataset, service.uri, dataset_url
                )
            )
            is not None
        ),
        KeyError(
            f"The requested dataset URL '{dataset_url}' "
            "was not found in the catalogue, "
            "you can use 'copernicus-marine describe --include-datasets "
            "--contains <search_token>' to find datasets"
        ),
    )


def _parse_dataset_and_suffix_url(
    dataset: CopernicusMarineProductDataset, service_uri: str, dataset_url: str
) -> Optional[Tuple[CopernicusMarineProductDataset, str]]:
    return (
        (dataset, dataset_url.split(service_uri)[1])
        if dataset_url.startswith(service_uri)
        else None
    )


def get_product_from_url(
    catalogue: CopernicusMarineCatalogue, dataset_url: str
) -> CopernicusMarineProduct:
    """
    Return the product object, with its dataset list filtered
    """
    filtered_catalogue = filter_catalogue_with_strings(
        catalogue, [dataset_url]
    )
    if filtered_catalogue is None:
        error = TypeError("filtered catalogue is empty")
        raise error
    if isinstance(filtered_catalogue, CopernicusMarineCatalogue):
        return filtered_catalogue.products[0]
    return filtered_catalogue["products"][0]


def filter_catalogue_with_strings(
    catalogue: CopernicusMarineCatalogue, tokens: list[str]
) -> dict[str, Any]:
    return find_match_object(catalogue, tokens) or {}


def find_match_object(value: Any, tokens: list[str]) -> Any:
    match: Any
    if isinstance(value, str):
        match = find_match_string(value, tokens)
    elif isinstance(value, Enum):
        match = find_match_enum(value, tokens)
    elif isinstance(value, tuple):
        match = find_match_tuple(value, tokens)
    elif isinstance(value, list):
        match = find_match_list(value, tokens)
    elif hasattr(value, "__dict__"):
        match = find_match_dict(value, tokens)
    else:
        match = None
    return match


def find_match_string(string: str, tokens: list[str]) -> Optional[str]:
    return string if any(token in string for token in tokens) else None


def find_match_enum(enum: Enum, tokens: list[str]) -> Any:
    return find_match_object(enum.value, tokens)


def find_match_tuple(tuple: Tuple, tokens: list[str]) -> Optional[list[Any]]:
    return find_match_list(list(tuple), tokens)


def find_match_list(object_list: list[Any], tokens) -> Optional[list[Any]]:
    def find_match(element: Any) -> Optional[Any]:
        return find_match_object(element, tokens)

    filtered_list: list[Any] = list(map_reject_none(find_match, object_list))
    return filtered_list if filtered_list else None


def find_match_dict(
    structure: dict[str, Any], tokens
) -> Optional[dict[str, Any]]:
    filtered_dict = {
        key: find_match_object(value, tokens)
        for key, value in structure.__dict__.items()
        if find_match_object(value, tokens)
    }

    found_match = any(filtered_dict.values())
    if found_match:
        new_dict = dict(structure.__dict__, **filtered_dict)
        structure.__dict__ = new_dict
    return structure if found_match else None



from typing import Callable, Union

import pandas
import xarray

from copernicus_marine_client.catalogue_parser.catalogue_parser import (
    CopernicusMarineDatasetServiceType,
    parse_catalogue,
)
from copernicus_marine_client.catalogue_parser.request_structure import (
    LoadRequest,
)
from copernicus_marine_client.core_functions.credentials_utils import (
    get_username_password,
)
from copernicus_marine_client.core_functions.services_utils import (
    CommandType,
    get_dataset_service_and_suffix_path,
)


def load_data_object_from_load_request(
    load_request: LoadRequest,
    disable_progress_bar: bool,
    arco_series_load_function: Callable,
    opendap_load_function: Callable,
) -> Union[xarray.Dataset, pandas.DataFrame]:
    catalogue = parse_catalogue(
        overwrite_metadata_cache=load_request.overwrite_metadata_cache,
        no_metadata_cache=load_request.no_metadata_cache,
        disable_progress_bar=disable_progress_bar,
    )
    dataset_service, _, _ = get_dataset_service_and_suffix_path(
        catalogue=catalogue,
        dataset_id=load_request.dataset_id,
        dataset_url=load_request.dataset_url,
        force_service_type=load_request.force_service,
        command_type=CommandType.LOAD,
        dataset_subset=load_request.get_time_and_geographical_subset(),
    )
    username, password = get_username_password(
        load_request.username,
        load_request.password,
        load_request.credentials_file,
    )
    load_request.dataset_url = dataset_service.uri
    if dataset_service.service_type in [
        CopernicusMarineDatasetServiceType.GEOSERIES,
        CopernicusMarineDatasetServiceType.TIMESERIES,
    ]:
        dataset = arco_series_load_function(
            username=username,
            password=password,
            dataset_url=load_request.dataset_url,
            variables=load_request.variables,
            geographical_parameters=load_request.geographical_parameters,
            temporal_parameters=load_request.temporal_parameters,
            depth_parameters=load_request.depth_parameters,
            chunks=None,
        )
    elif (
        dataset_service.service_type
        == CopernicusMarineDatasetServiceType.OPENDAP
    ):
        dataset, _ = opendap_load_function(
            username=username,
            password=password,
            dataset_url=load_request.dataset_url,
            variables=load_request.variables,
            geographical_parameters=load_request.geographical_parameters,
            temporal_parameters=load_request.temporal_parameters,
            depth_parameters=load_request.depth_parameters,
        )
    return dataset
